@@ -1,25 +1,27 @@
-#' Natural-Logarithm-Transformed Odds Ratio (lnOR): Bootstrap-Based Heterogeneity Test for Between-study Heterogeneity in Random- or Mixed- Effects Model
+#' Natural-Logarithm-Transformed Odds Ratio (lnOR): Bootstrap-based Heterogeneity Test for Between-study Heterogeneity in Random- or Mixed- Effects Model
 #'
 #' \code{boot.lnOR} returns the bootstrap-based tests of the residual heterogeneity in random- or mixed- effects model of natural-logarithm-transformed observed odds ratio (lnOR).
 #'
 #' For odds ratio, its standard error will be infinite if any one of the four cells in the contingency tables is zero. In this case, Haldane and Anscombe correction is used by adding 0.5 to each cell value (Anscombe, 1956; Haldane, 1940).
 
-#' This function returns the test statistics as well as their p-value and significances using (1) Q-test, (2) Bootstrap-Based Heterogeneity Test with Maximum Likelihood (ML), and (3) Bootstrap-Based Heterogeneity Test with Restricted Maximum Likelihood (REML).
+#' This function returns the test statistics as well as their p-value and significances using (1) Q-test and (2) Bootstrap-based Heterogeneity Test with Restricted Maximum Likelihood (REML).
 #'
 #' The results of significances are classified as "sig" or "n.s" based on the cutoff p-value (i.e., alpha level). "sig" means that the between-study heterogeneity is significantly different from zero whereas "n.s" means the between-study heterogeneity is not significantly different from zero. The default alpha level is 0.05.
 #'
-#' @param n_00 a vector of number of participants who score negatively on both Y1 and Y2 (e.g., mortality cases in the control group).
-#' @param n_01 a vector of number of participants who score negatively on Y1 and positively on Y2  (e.g., recovery cases in the control group).
-#' @param n_10 a vector of number of participants who score positively on Y1 and negatively on Y2 (e.g., mortality cases in the experimental group).
-#' @param n_11 a vector of number of participants who score positively on both Y1 and Y2 (e.g., recovery cases in the experimental group).
-#' @param lnOR a vector of natural-logarithm-transformed odds ratio in the included studies, which is calculated as ln(n11*n00/n01/n10)
-#' @param model choice of random- or mixed- effects models. Can only be set to \code{"random"}, or \code{"mixed"}.
-#' @param mods optional argument to include one or more moderators in the model. \code{mods} is NULL for random-effects model and a dataframe for mixed-effects model. A single moderator can be given as a vector of length \eqn{k} specifying the values of the moderator. Multiple moderators are specified by giving a matrix with \eqn{k} rows and as many columns as there are moderator variables. See \code{\link[metafor:rma.uni]{rma}} for more details.
-#' @param nrep number of replications used in bootstrap simulations. Default to 10^4.
-#' @param p_cut cutoff for p-values, which is the alpha level. Default to 0.05.
-#' @param boot.include if true, bootstrap simulation results are included in the output (e.g., bootstrap critical values).
-#' @param parallel if true, parallel computing using 2 cores will be performed during bootstrapping stage. Otherwise, for loop is used.
-#' @param verbose if true, show the progress of boostrapping.
+#' @param n_00 A vector of number of participants who score negatively on both Y1 and Y2 (e.g., mortality cases in the control group).
+#' @param n_01 A vector of number of participants who score negatively on Y1 and positively on Y2  (e.g., recovery cases in the control group).
+#' @param n_10 A vector of number of participants who score positively on Y1 and negatively on Y2 (e.g., mortality cases in the experimental group).
+#' @param n_11 A vector of number of participants who score positively on both Y1 and Y2 (e.g., recovery cases in the experimental group).
+#' @param lnOR A vector of natural-logarithm-transformed odds ratio in the included studies, which is calculated as ln(n11*n00/n01/n10)
+#' @param lambda Size of the magnitude to be tested in the alternative hypothesis of the heterogeneity magnitude test. Default to 0.
+#' @param model Choice of random- or mixed- effects models. Can only be set to \code{"random"}, or \code{"mixed"}.
+#' @param mods Optional argument to include moderators in the model. \code{mods} is NULL for random-effects model and a dataframe of moderators for mixed-effects model. A single moderator can be given as a vector specifying the values of the moderator. Multiple moderators are specified by giving a matrix with as many columns as there are moderator variables. See \code{\link[metafor:rma.uni]{rma}} for more details.
+#' @param nrep Number of replications used in bootstrap simulations. Default to 10^4.
+#' @param p_cut Cutoff for p-value, which is the alpha level. Default to 0.05.
+#' @param boot.include If true, bootstrap simulation results are included in the output (e.g., bootstrap critical values).
+#' @param parallel If true, parallel computing using 4 cores will be performed during bootstrapping stage. Otherwise, for loop is used.
+#' @param cores The number of cores used in the parallel computing. Default to 4.
+#' @param verbose If true, show the progress of bootstrapping.
 #'
 #' @return A dataframe that contains the test statistics ('stat'), p-values ('p_value'), and significances of effect size heterogeneity ("Heterogeneity").
 #'
@@ -27,6 +29,9 @@
 #' @importFrom metafor fitstats
 #' @importFrom pbmcapply pbmclapply
 #' @importFrom stats runif
+#' @importFrom stats rnorm
+#' @importFrom stats na.omit
+#' @importFrom stats ecdf
 
 #' @references Anscombe, F. J. (1956). On estimating binomial response relations. Biometrika, 43(3/4), 461–464.
 #' @references Haldane, J. (1940). The mean and variance of| chi 2, when used as a test of homogeneity, when expectations are small. Biometrika, 31(3/4), 346–355.
@@ -50,7 +55,7 @@
 #' }
 #' @export
 
-boot.lnOR <- function(n_00, n_01, n_10, n_11, model = 'random', mods = NULL, nrep = 10^4, p_cut = 0.05, boot.include = FALSE, parallel = FALSE, verbose = FALSE) {
+boot.lnOR <- function(n_00, n_01, n_10, n_11, lambda = 0, model = 'random', mods = NULL, nrep = 10^4, p_cut = 0.05, boot.include = FALSE, parallel = FALSE, cores = 4, verbose = FALSE) {
 
   #########################################################################
   if (!model %in% c('random', 'mixed')){
@@ -79,8 +84,8 @@ boot.lnOR <- function(n_00, n_01, n_10, n_11, model = 'random', mods = NULL, nre
   lnOR <- log(n_11*n_00/n_01/n_10)
   vi <- 1/n_00+1/n_01+1/n_10+1/n_11
 
-  model.f1<-try(metafor::rma(lnOR, vi, mods = mods, tau2=0, method="ML"))
-  model.f2<-try(metafor::rma(lnOR, vi, mods = mods, tau2=0, method="REML"))
+  model.f1<-try(metafor::rma(lnOR, vi, mods = mods, tau2=lambda^2, method="ML"))
+  model.f2<-try(metafor::rma(lnOR, vi, mods = mods, tau2=lambda^2, method="REML"))
   model.r1<-try(metafor::rma(lnOR, vi, mods = mods, method="ML"))
   model.r2<-try(metafor::rma(lnOR, vi, mods = mods, method="REML"))
 
@@ -93,7 +98,7 @@ boot.lnOR <- function(n_00, n_01, n_10, n_11, model = 'random', mods = NULL, nre
   if(verbose){cat("Bootstrapping... \n")}
 
   if(parallel){
-    find.c <- do.call(cbind, pbmcapply::pbmclapply(1:nrep, simulate.OR, lnOR_overall = lnOR_overall, vi = vi, n = n, n_00_s = n_00, n_01_s = n_01, n_10_s = n_10, n_11_s = n_11, mods = mods, mc.cores = 2))
+    find.c <- do.call(cbind, pbmcapply::pbmclapply(1:nrep, simulate.OR, lnOR_overall = lnOR_overall, lambda=lambda, vi = vi, n = n, n_00_s = n_00, n_01_s = n_01, n_10_s = n_10, n_11_s = n_11, mods = mods, mc.cores = cores))
     # parallel::detectCores()-1)
   } else {
     find.c <- matrix(NA, 3, nrep)
@@ -101,7 +106,7 @@ boot.lnOR <- function(n_00, n_01, n_10, n_11, model = 'random', mods = NULL, nre
     for(i in 1:nrep){
       Sys.sleep(0.01)
       utils::setTxtProgressBar(pb, i)
-      find.c[,i] = simulate.OR(i, lnOR_overall, vi, n, n_00, n_01, n_10, n_11, mods)
+      find.c[,i] = simulate.OR(i, lnOR_overall, lambda, vi, n, n_00, n_01, n_10, n_11, mods)
     }
   }
   err.catcher <- sum(colSums(is.na(find.c))!=0)/nrep
@@ -109,7 +114,15 @@ boot.lnOR <- function(n_00, n_01, n_10, n_11, model = 'random', mods = NULL, nre
     warning("Noncovergence rate in simulations is larger than 5%!")
   }
 
-  ML.sim <- stats::na.omit(unlist(find.c)[ c(TRUE,FALSE,FALSE) ])
+  # We recommend B-REML-LR
+  # p-value
+  if (model.r1$tau2>=(lambda^2)){
+    # One-sided test so the estimated tau has to be larger than lambda,
+    # otherwise, we fail to reject the null hypothesis.
+    #f <- ecdf(na.omit(unlist(find.c)[ c(FALSE,TRUE,FALSE) ]))
+    #pvalue=1-f( (fitstats(model.r2)-fitstats( model.f2))[1]*2)
+    # Ge's way to calculate p-value (it's the same)
+    ML.sim <- stats::na.omit(unlist(find.c)[ c(TRUE,FALSE,FALSE) ])
   REML.sim <- stats::na.omit(unlist(find.c)[ c(FALSE,TRUE,FALSE) ])
   chisq.sim <- stats::na.omit(unlist(find.c)[ c(FALSE,FALSE,TRUE) ])
   ML.c<-stats::quantile(ML.sim, 0.95)
@@ -139,6 +152,11 @@ boot.lnOR <- function(n_00, n_01, n_10, n_11, model = 'random', mods = NULL, nre
   Q <- model.f1$QE
   Qp <- model.r2$QEp
   Qres<-ifelse(Qp<= p_cut, 'sig', 'n.s') ### vary the size
+  } else {
+    pvalue=NA
+    warning("pvalue=NA and we fail to reject the null hypothesis.")
+    #We don't calculate pvalue in this case and just say that we fail to reject the null hypothesis.
+  }
   } else {
     Q<-NA
     Qp<-NA
@@ -160,8 +178,13 @@ boot.lnOR <- function(n_00, n_01, n_10, n_11, model = 'random', mods = NULL, nre
   #out <- data.frame(stat = c(Q, Q, lllr1, lllr2), p_value = c(Qp, p_Q, p_lr1, p_lr2), Heterogeneity = c(Qres, res_bootQ, res_lr1, res_lr2))
   #rownames(out) <- c('Qtest', 'boot.Qtest', 'boot.ML', 'boot.REML')
 
-  out <- data.frame(stat = c(Q, Q, lllr2), p_value = c(Qp, p_Q, p_lr2), Heterogeneity = c(Qres, res_bootQ, res_lr2))
-  rownames(out) <- c('Qtest', 'boot.Qtest', 'boot.REML')
+  if(lambda==0){
+    out <- data.frame(stat = c(Q, lllr2), p_value = c(Qp, p_lr2), Heterogeneity = c(Qres, res_lr2))
+    rownames(out) <- c('Qtest', 'boot.REML')
+  } else {
+    out <- data.frame(stat = c(lllr2), p_value = c(p_lr2), Heterogeneity = c(res_lr2))
+    rownames(out) <- c('boot.REML')
+  }
 
   if(boot.include){
     out <- list(results = out, find.c = find.c, ML.sim = ML.sim, REML.sim = REML.sim, chisq.sim = chisq.sim)
